--------------------------------------------------------------------------------
--
--  Murphi Model of the Bitcoin protocol
--
--------------------------------------------------------------------------------
--
--  version:      1.0
--
--  written by:   Ruven Chu & Andrew He
--  date:         Mar 2011
--  affiliation:  Stanford University (graduate student)
--
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- constants, types and variables
--------------------------------------------------------------------------------


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
const
  INTRUDERACTION:     TRUE;   -- Should the intruder be able to do special actions?
  NumParticipators:   2;      -- normal participators
  NumIntruders:       1;      -- intruders (when INTRUDERACTION is false, Intruder = Participator)
  TransCount:         2;      -- number of total transactions
  StartCoins:         1;      -- number of coins each agent starts with
  MaxCoins:           10;     -- max number of coins that can be owned
  ParticipatorWork:   3;      -- how much work a normal participant has
  IntruderWork:       3;      -- how much work an intruder has
  MaxWork:            3;      -- max amount of work a user can have


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
type
  TransId:   0..TransCount;                        -- Transaction ID
  ValidTransId: 1..TransCount;                     -- Valid transaction ID (0 is not a real transaction)
  ParticipatorId:  scalarset (NumParticipators);   -- ParticipatorID
  IntruderId:   scalarset (NumIntruders);          -- IntruderID
  CurrentTrans: 1..TransCount+1;                   -- Range of global transactions
  WorkAmount: 0..MaxWork;                          -- Amount of work someone can have
  TotalCoins: -MaxCoins..MaxCoins;                 -- Amount of coins someone can have (used to be we could have negative amounts)
  
  AgentId:      union {ParticipatorId, IntruderId};
  
  -- either calculating something or siting around waiting for a new transaction
  Status : enum {
    Calculating,
    Waiting
  };
  
  -- Coin "from and two" is the source and test, ID and prev (to make a tree)
  Block : record
    source: AgentId;
    dest:   AgentId;
    id:     ValidTransId;
    prev:   TransId;
  end;
  
  BlockChain : array[ValidTransId] of Block;          -- Total list of blocks/transactions
  ValidChain : multiset[TransCount] of ValidTransId;  -- multiset of valid transaction-ids
  CoinCounts : array[AgentId] of TotalCoins;          -- Counts of coins
  
  -- Participators and Intruders can do the same things
  Agent : record
    chain: BlockChain;        -- Total chain it knows (for more efficient space, we should have just kept IDs)
    counts: CoinCounts;       -- Current coin counts for valid chain
    validchain: ValidChain;   -- IDs of valid chain
    working: Block;           -- Block it's currently working on
    status: Status;           -- Status of the Agent
    workleft: WorkAmount;     -- How much work this agent has left (simulating processing power)
  end;
    

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var                                         -- state variables for
  age: array[AgentId] of Agent;             -- all agents
  cur: CurrentTrans;                        -- current transaction counter
  tra: BlockChain;                          -- global transaction list (no need for agent specific ones, oh well)

--------------------------------------------------------------------------------
-- helper functions
--------------------------------------------------------------------------------

-- Get's the counts based on the longest valid chain
function getcounts(var longest: ValidChain) : CoinCounts;
var counts : CoinCounts;
var toc: TotalCoins;
var fromc: TotalCoins;
begin
  for i: AgentId do
    toc := MultisetCount(j:longest, tra[longest[j]].dest = i);      -- number of coins given
    fromc := MultisetCount(j:longest, tra[longest[j]].source = i);  -- number of coins given up
    counts[i] := StartCoins + toc - fromc;
  end;
  return counts;
end;

-- Checks to make sure no agent double pays - combines all transactions from valid chains
-- from both, and then checks to see if anyone has negative
function checknodoublepay(one: Agent; two: Agent) : boolean;
var combined: ValidChain;
var counts: CoinCounts;
begin
  undefine combined;
  for i: ValidTransId do
    -- if the transaction is in either chain, we add it to the temporary chain
    if MultisetCount(j:one.validchain,one.validchain[j] = i) > 0 | MultisetCount(j:two.validchain,two.validchain[j] = i) > 0  then
      MultisetAdd(i, combined);
    end;
  end;
  counts := getcounts(combined);
  for i:AgentId do
    if counts[i] < 0 then  -- if anyone has negative - double spending going on
      return false;
    end;
  end;
  return true;
end;

-- This checks to see if a transaction can be added to a user's total chain - if it creates a 
-- negative cointcount it's not good. Only intruders can send these transactions
function getvalid(lasttrans: TransId) : boolean;
var tmp: TransId;
var tmpchain: ValidChain;
var counts : CoinCounts;
begin
  undefine tmpchain;
  tmp := lasttrans;
  -- follow chain back, get chain length
  while tmp > 0 do
    MultisetAdd(tra[tmp].id, tmpchain);
    tmp := tra[tmp].prev;
  end;
  
  -- check to see if illegal counts
  counts := getcounts(tmpchain);
  for i:AgentId do
    if counts[i] < 0 then
      return false;
    end;
  end;
return true;
end;

-- Get's the longest chain given a user's current blockchain
-- recursive
-- deterministic given the same input
procedure getlongest(var input:BlockChain; var current, biggest: ValidChain; last : 0..TransCount);
var tmp:Block;
var hit:boolean;
begin
  hit := false;
  for i: ValidTransId do   
    if(!isundefined(input[i].prev) & input[i].prev = last) then
      hit := true;
      tmp := input[i];
      -- add the new block to the current chain
      MultisetAdd(input[i].id, current);
      undefine input[i];
      getlongest(input, current, biggest, tmp.id);
      -- then remove it after the recursive call
      MultisetRemovePred(j:current, current[j] = tmp.id);
      input[i] := tmp;
    end;
  end;
  -- if no blocks were added then we've hit a long chain, check if we should replace
  if hit = false & MultiSetCount(k:current, k=k) > MultiSetCount(k:biggest, k=k) then
    biggest := current;
  end;
end;

-- Copies an entire chain backwards into someone's input (if imperfect boardcast,
-- could be missing previous blocks)
procedure copychain(var input:BlockChain; var trans:Block);
var cur:TransId;
begin
  cur := trans.id;
  while cur > 0 do
    if isundefined(input[cur].id) then
      input[cur] := tra[cur];   
    end;
    cur := tra[cur].prev;
  end;
end;

--------------------------------------------------------------------------------
-- rules
--------------------------------------------------------------------------------

-- starting the transaction, special case because prev is 0
ruleset i: AgentId do
  ruleset l: AgentId do
    rule 12 "participator sends 1st transaction"
    
      -- sends a transaction to someone else
      cur = 1 &
      age[i].status = Waiting &
      i != l  &
      age[i].counts[i] > 0
    ==>
    
    var
      outM: Block;   -- outgoing Block

    begin
      undefine outM;
      outM.source  := i;
      outM.dest   := l;
      outM.prev    := 0;
      outM.id      := cur;
      
      -- goes through and sets everyone to working on this
      for j: AgentId do   
        if isundefined(age[j].working.id) then
          age[j].working := outM;
          age[j].status := Calculating;
        end;
      end;
      
      -- sets the global array
      tra[cur] := outM;
      cur := cur + 1;
    end;
  end;
end;

-- intruder can selectively send the first transaction
ruleset i: IntruderId do
  ruleset l: AgentId do
    rule 4 "intruder sends 1st transaction to only one agent"
    
      cur = 1 &
      age[i].status = Waiting &
      i != l  &
      INTRUDERACTION
    ==>
    
    var
      outM: Block;   -- outgoing Block

    begin
      undefine outM;
      outM.source  := i;
      outM.dest   := l;
      outM.prev    := 0;
      outM.id      := cur;
       
      -- only to one agent
      if isundefined(age[l].working.id) then
        age[l].working := outM;
        age[l].status := Calculating;
      end;
      
      age[i].working := outM;
      age[i].status := Calculating;
      
      tra[cur] := outM;
      cur := cur + 1;
    end;
  end;
end;

-- participant adds another transaction, sets a from, to, and a previous transaction id
ruleset i: AgentId do
  ruleset l: AgentId do
    ruleset k: TransId do
      rule 15 "participator adds another transaction"
      
        cur <= TransCount &
        age[i].status = Waiting &
        k < cur &
        i != l &
        age[i].counts[i] > 0 &
        (k = 0 | !isundefined(age[i].chain[k].id)) -- can only add another transaciton that makes sense to it

      ==>
    
      var
        outM: Block;   -- outgoing Block

      begin
        undefine outM;
        outM.source  := i;
        outM.dest   := l;
        outM.prev    := k;
        outM.id      := cur;
     
        for j: AgentId do   
          if age[j].status = Waiting then
            age[j].working := outM;
            age[j].status := Calculating;
          end;
        end;
        
        tra[cur] := outM;
        cur := cur + 1;
      end;
    end;
  end;
end;

-- intruder can selectively send a transaction to someone
ruleset i: IntruderId do
  ruleset l: AgentId do
    ruleset k: TransId do
      rule 5 "intruder adds another transaction"
      
        cur <= TransCount &
        age[i].status = Waiting &
        k < cur &
        i != l &
        INTRUDERACTION

      ==>
    
      var
        outM: Block;   -- outgoing Block

      begin
        undefine outM;
        outM.source  := i;
        outM.dest   := l;
        outM.prev    := k;
        outM.id      := cur;
     
        if age[l].status = Waiting then
          age[l].working := outM;
          age[l].status := Calculating;
        end;
        
        age[i].working := outM;
        age[i].status := Calculating;
        
        tra[cur] := outM;
        cur := cur + 1;
      end;
    end;
  end;
end;

-- someone finishes working on a transaction, commits to everyone
ruleset i: AgentId do
  rule 20 "someone commits the transaction"
    
    age[i].status = Calculating &
    age[i].workleft > 0
    
    ==>
    var
      tmpworking: Block;
      tmpchain: ValidChain;
      bigchain: ValidChain;
      chaincopy: BlockChain;
      tmpcounts: CoinCounts; 
    begin

      tmpworking := age[i].working;            -- gets the block it's working on
      for j: AgentId do   
        chaincopy := age[j].chain;             -- makes a copy of the chain
        copychain(chaincopy, tmpworking);      -- copies the working block into the chian
        if getvalid(tmpworking.id) then        -- if this is a valid addition
          undefine tmpchain;
          undefine bigchain;
          getlongest(chaincopy, tmpchain, bigchain, 0);  -- get's the new longest valid chain
          tmpcounts := getcounts(bigchain);              -- get's the new conuts
          age[j].validchain := bigchain;                 -- copies in the long chain
          age[j].chain := chaincopy;                     -- copies in the new blcoks
          age[j].counts := tmpcounts;                    -- copies in the counts
        end;
        -- only if it was working on this we stop what it was working on
        if !isundefined(age[j].working.id) & age[j].working.id = tmpworking.id then
          undefine age[j].working;
          age[j].status := Waiting;
        end;
      end;
      -- uses up some work
      age[i].workleft := age[i].workleft - 1;
  end;
end;

-- same as above - except only commits to one person
ruleset i: IntruderId do
  rule 10 "intruder selectively commits a transactions"
    
    age[i].status = Calculating &
    age[i].workleft > 0 & 
    age[i].working.source = i &  -- for simplicity, only commits own transactions as illegal
    INTRUDERACTION
    
  ==>
  var
    tmpworking: Block;
    tmpchain: ValidChain;
    bigchain: ValidChain;
    chaincopy: BlockChain;
    tmpcounts: CoinCounts;
    otherid: AgentId;
  
  begin 
    tmpworking := age[i].working;
    otherid := age[i].working.dest; -- for simplicity again, only commits to the person it had sent this transaction to
  
    chaincopy := age[i].chain;
    copychain(chaincopy, tmpworking);
    if getvalid(tmpworking.id) then
      undefine tmpchain;
      undefine bigchain;
      getlongest(chaincopy, tmpchain, bigchain, 0);
      tmpcounts := getcounts(bigchain);
      age[i].validchain := bigchain;
      age[i].chain := chaincopy;
      age[i].counts := tmpcounts;
    end;
    undefine age[i].working;
    age[i].status := Waiting;
    
    chaincopy := age[otherid].chain;
    copychain(chaincopy, tmpworking);
    if getvalid(tmpworking.id) then
      undefine tmpchain;
      undefine bigchain;
      getlongest(chaincopy, tmpchain, bigchain, 0);
      tmpcounts := getcounts(bigchain);
      age[otherid].validchain := bigchain;
      age[otherid].chain := chaincopy;
      age[otherid].counts := tmpcounts;
    end;
    if !isundefined(age[otherid].working.id) & age[otherid].working.id = tmpworking.id then
      undefine age[otherid].working;
      age[otherid].status := Waiting;
    end;
      
    age[i].workleft := age[i].workleft - 1;
  end;
end;

--------------------------------------------------------------------------------
-- startstate
--------------------------------------------------------------------------------

startstate
  -- initialize peops
  undefine age;
  for i: ParticipatorId do
    age[i].status := Waiting;
    age[i].workleft := ParticipatorWork;
    age[i].counts := getcounts(age[i].validchain);
  end;
  
  for i: IntruderId do
    age[i].status := Waiting;
    age[i].workleft := IntruderWork;
    age[i].counts := getcounts(age[i].validchain);
  end;
 
  cur := 1;
end;

--------------------------------------------------------------------------------
-- invariants
--------------------------------------------------------------------------------

-- check double-pay invariant, fails when intruder can make special actions
invariant "nobody legally double pays"
  forall i: ParticipatorId do
    forall j: ParticipatorId do
      i != j
      ->
      checknodoublepay(age[i],age[j])
    end
  end;

  
